# Potatoe offers a new way to define objects
# in CoffeeScript.
# It is especially suited for objects that rely
# heavily on composition (e.g. : Models).
#

extend = (dest,extra) ->
    for k,v of extra
        dest[k] = v
    dest

recursiveExtend = (dest, extra)->
    for k,v of extra
        if (typeof v) == "object"
            if not dest[k]?
                dest[k] = {}
            recursiveExtend dest[k], v
        else
            dest[k]=v
    dest

#
# **Tuber** is the root of potatoe inheritance.
# 
# Potatoe inheritance uses an unusual syntax.
# A Potatoe itself is the function allowing to extend
# itself.
#
# Instead of going :
#
#     class Animal
#         # Definition of the content of the 
#         # Animal class.
#           
#     class Dog extends Animal
#         # Definition of the content of the 
#         # Dog class. Dogs inherits from
#         # Animal
# 
# With Potatoe, you would write:
# 
#     Animal = Potatoe
#         # Definition of the content of the 
#         # Animal class.
#
#     Dog = Animal
#         # Definition of the content of the 
#         # Dog class. Dogs inherits from
#         # Animal
#
# While the purpose of such a syntax may 
# not make sense at the moment,
# we will see that this syntax makes it possible to
# define anonymous classes, which is precious when
# programming while relying heavily on composition
# and deep overriding.
#
# Tuber also makes it possible for someone 
# to define static members for your class.
# 
Tuber = (content)->

    if not content.static?
        content.static = {}
    for k,v of content.methods
        do (k,v) ->
            if not content.static[k]?
                content.static[k] = (self,args...)->
                    content.methods[k].call self, args...
    
    extendMyself = (extraContent)->
        # For each method, automatically add pythonic
        # methods to the class.
        if @constructor.THIS_IS_NOT_A_CONSTRUCTOR_DUMMY?
            msg = "Do no call 'new YourPotato()'. Instanciation is done via YourPotato.make()'."
            throw msg
        new_content = recursiveExtend {}, content
        recursiveExtend new_content, extraContent
        Tuber new_content
    
    extendMyself.THIS_IS_NOT_A_CONSTRUCTOR_DUMMY = true
    recursiveExtend extendMyself,
        content.static
    
    extendMyself.__content = content
    extendMyself

Tuber.__content = {}


# ---------------------------
# **Potato** is the real deal.
# 
# You may have notice that Tubers cannot be instanciated.
# The inheritance syntax conflicts with the usual
# instanciation syntax. We are not able to do 
#
#     dog = new Dog()
#
# Instanciation will be done via a "make" static
# method. Which basically delegates to its components.
#

delegateTo = (delegateMember, methodName)->
    ->
        member = (delegateMember.apply this)
        member[methodName].apply member, arguments

Potato = Tuber
    
    methods:
        components: ->
            @__potato__.components this
        
        set: (data)->
            components = @components()
            for k,v of data
                component = components[k]
                if component?
                    this[k] = component.set this[k],v
                else
                    Potato.error k + " is not a member of this potato."
            this
        
        # Returns JSON String describing the state.
        toJSON: ->
            JSON.stringify @toData()
        
        # Returns JSON Object describing the state
        toData: ->
            data = {}
            for k,v of @components()
                data[k] = v.toData this[k]
            data
        
    static:
        __init__: (obj)->
            
        # Produce an instance with 
        # the data "data"
        make: (data={})->
            potato = this
            actualConstructor = (data={})->
                for k,v of potato.__content.components
                    this[k] = v.make()
                this
            actualConstructor::__potato__ = potato
            extend actualConstructor.prototype, potato.__content.methods

            # You may delegate some method to some
            # of the member of your potato.
            for methodName,delegateMember of potato.__content.delegate
                actualConstructor.prototype[methodName] = delegateTo delegateMember, methodName
            newInstance = new actualConstructor
            newInstance.set data
            @__init__ newInstance
            newInstance
        


        
        # Returns an object from this JSON data.        
        fromJSON: (json)->
            data = JSON.parse json
            @fromData data
        
        fromData: (data)->
            @make data
        
        components: ->
            @__content.components


Potato.error = (args...)->
    for arg in args
        console.log "ERROR : ", arg



# ---------------------------
# Potato skin is for complex
# models which do not have any 
# components.
#
# All its data is stored in @data
#
Potato.Skin = Potato

    static:
        set: (obj, data)->
            obj.data = data
            obj
        
        toData: (obj)->
            obj.data

# ---------------------------
# You need some kind of Leaf to this tree of composed models.
# That's the role of Potato.Litteral. 
# 
# Litteral may not contain any methods. They 
# are actual javascript litterals, which 
# allows you to get/set those via simple
# 
# Serialization, etc. can be done via the static
# method of their Tuber.
#
#     dog.age = 12
#
#
Potato.Litteral = Tuber
    
    static:
        make: (val)->
            if val?
                val
            else
                @__content.default
        toJSON: (val)->
            JSON.stringify @toData val
        toData: (val)->
            val
        set: (obj,val)->
            val


# ----------------------------------
# List are also handled by Potato
#
Potato.List = Potato
    
    methods:
        push: (item)->
            @items.push item
        get: (i)->
            @items[i]
        toData: ->
            @__potato__.itemType.toData it for it in @items        
        set: (itemsData)->
            @items = ( (@__potato__.itemType.make itemData) for itemData in itemsData )
            this
    
    static:
        itemType: Potato.Litteral
            default: null
        
        make: (data=[])->
            potato = this
            actualConstructor = ->
                @items = []
                this
            actualConstructor.prototype.__potato__ = potato
            extend actualConstructor.prototype, potato.__content.methods
            newInstance = new actualConstructor()
            newInstance.set data
            newInstance


Potato.ListOf = (itemType)->
    Potato.List 
        static:
            itemType: itemType




# ----------------------------------
# ... as well as maps
#
Potato.Map = Potato
    
    methods:
        push: (key, value)->
            @map[key] = value
        
        get: (key)->
            @map[key]
        
        toData: ->
            data = {}
            for k,v of @map
                data[k] = @__potato__.itemType.toData v
            data
        
        set: (data)->
            # TODO consider a 1-level lower "override behavior"
            @map = {}
            for k,v of data
                @push k, @__potato__.itemType.make v
            this
        
    static:
        itemType: Potato.Litteral
            default: null
        make: (data={})->
            potato = this
            actualConstructor = ->
                @map = {}
                this
            actualConstructor.prototype.__potato__ = potato
            extend actualConstructor.prototype, potato.__content.methods
            newInstance = new actualConstructor()
            newInstance.set data
            newInstance

Potato.MapOf = (itemType)->
    Potato.Map
        static:
            itemType: itemType

Potato.runTests = (TESTS)->
    for testId in [ 0 ... TESTS.length ]
        try
            TESTS[testId]()
            console.log '[  OK  ]', '   Test' , testId + 1,
        catch e
            console.log e
            console.log '[ FAIL ]', '   Test' , testId + 1


apply_mixins = (potato, mixins)->
    for mixin in mixins
        potato = potato(mixin)
    potato

Potato.mixin = (mixins...)->
    PotatoWithMixins = apply_mixins this, mixins
    for k,v of this
        if v.__content? and k != "Mixin"
            PotatoWithMixins[k] = apply_mixins this[k],mixins
    PotatoWithMixins

Potato.Mixin = (x)->x

Potato.notImplemented = ->
    throw "Not Implemented Error"



# Exports Potato for both NodeJS
# and the browser ...
if window?
    window.Potato = Potato
else
    module.exports = Potato
  